#!/bin/bash

OS_ID=`. /etc/os-release && echo $ID`
OS_VERSION_ID=`. /etc/os-release && echo $VERSION_ID`
OS_NAME=$OS_ID$OS_VERSION_ID
CHECK_ARCH=`arch` 
KERNEL_VERSION=`uname -r`
GENBU_SERVICE="/lib/systemd/system/genbu.service"
INITRD_FILE="/boot/initrd.img-${KERNEL_VERSION}"

function install_gb_config() {
if [ "$OS_NAME" == "kylinv4" ]; then
    cp /usr/local/genbu/rules/91-pulseaudio-custom.rules /lib/udev/rules.d/ -rf
else
    cp /usr/local/genbu/rules/91-pulseaudio-custom.rules /usr/lib/udev/rules.d/ -rf
fi
}

function update_sietium_pciids() {
    # export Sietium's pci.ids in system file(If exist)
    sed -n "/Sietium/,$ {
    /Sietium/ {
        :loop
        p
        n
        /\t+*/ {
            b loop
        }
    }
    }" $1 > pci.sietium.ids.old

        # export Sietium's pci.ids in local file(If exist)
        sed -n "/Sietium/,$ {
    /Sietium/ {
        :loop
        p
        n
        /\t+*/ {
            b loop
        }
    }
    }" /usr/local/sysconf/pci.ids > pci.sietium.ids.new

    # compare file exclude white space
    if diff -B pci.sietium.ids.new pci.sietium.ids.old&>/dev/null; then
        # echo "sietium pci.ids no update required. >> ESC"
                rm pci.sietium.ids.old
                rm pci.sietium.ids.new
    else
        echo "sietium pci.ids update required. >> UPDATE"
        # remove exist Sietium's pci.ids in system pci.ids file
        sed "/8510  Sietium/,$ {
            /8510  Sietium/d
            :loop
            /\t+*/ {
                d
                b loop
            }
            :loop {
                n
                b loop
            }
        }
        " $1 > pci.ids.system

        # append new line at end of pci.ids.system file
        if [ $(tail -n1 pci.ids.system | wc -l) -eq 0 ];then
            echo "" >> pci.ids.system
        fi

        # append new Sietium's pci.ids to pci.ids file
        cat pci.sietium.ids.new >> pci.ids.system
        # replace pci.ids file
        mv pci.ids.system $1
                rm pci.sietium.ids.old
                rm pci.sietium.ids.new
    fi
}

update_sietium_pciids /usr/share/misc/pci.ids

if [ "$OS_NAME" == "kylinv4" ]; then
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/hdmi_sound_toggle.sh /usr/local/bin/hdmi_sound_toggle.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/PCIE_FULL_DDR4_sound.sh /usr/local/bin/PCIE_FULL_DDR4_sound.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/PCIE_FULL_LPDDR4_sound.sh /usr/local/bin/PCIE_FULL_LPDDR4_sound.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/PCIE_LPDDR4_sound.sh /usr/local/bin/PCIE_LPDDR4_sound.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/PCIE_M6FL8G_LPDDR4_sound.sh /usr/local/bin/PCIE_M6FL8G_LPDDR4_sound.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/PCIE_HIE1LP4_LPDDR4_sound.sh /usr/local/bin/PCIE_HIE1LP4_LPDDR4_sound.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/PCIE_CQ2040_MXM_M60_sound.sh /usr/local/bin/PCIE_CQ2040_MXM_M60_sound.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/99-hdmi_sound.rules /etc/udev/rules.d/99-z_hdmi_sound.rules
    #cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/asound.conf /etc/asound.conf
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/z-snd-autostart.desktop /etc/xdg/autostart/z-snd-autostart.desktop
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/genbu-script.sh /usr/local/bin/genbu-script.sh
    cp /usr/src/gb02-1.1.600-2/scripts/hdmi_files/genbu-execute.sh /usr/local/bin/genbu-execute.sh
fi
#find "/usr/src/gb02-1.1.600-2" -name ".git*" -exec rm -rf {} +

install_gb_config
udevadm control --reload-rules


if grep -q "kylin" /etc/os-release; then
    . /etc/default/grub
    GRUB_DEFAULT=`echo $GRUB_CMDLINE_LINUX_DEFAULT | sed 's/video=efifb:on//g'`
    sed -i "/GRUB_CMDLINE_LINUX_DEFAULT/s/=.*/=\"${GRUB_DEFAULT} video=efifb:on\"/" /etc/default/grub
    if test "$CHECK_ARCH" == "aarch64"; then
        update-grub-ftefi
    fi
    if [ -e $GENBU_SERVICE ];then
        systemctl enable -q $GENBU_SERVICE
    fi
    grep -qE '^gb02$|^gb02\s|^gb02\t' /etc/initramfs-tools/modules ||         echo "gb02" >> /etc/initramfs-tools/modules
    echo "KYLIN install update-initramfs......"
    update-initramfs -u -k $KERNEL_VERSION >/dev/null 2>&1
else
    CHECK_UOS="uos"
    UOS_VER="/etc/os-release"
    if [ -e $UOS_VER ];then
        result=$(cat $UOS_VER|grep $CHECK_UOS)
        if test -n "$result"; then
            cp /usr/local/sysconf/genbu.conf /etc/modules-load.d/genbu.conf
            grep -qE '^gb02$|^gb02\s|^gb02\t' /etc/initramfs-tools/modules ||                 echo "gb02" >> /etc/initramfs-tools/modules
            echo "update-initramfs......"
            update-initramfs -u -k $KERNEL_VERSION >/dev/null 2>&1
        fi
    fi
fi


#update mcu firmware
chmod 700 /usr/local/sietium/mcu_upgrade.sh
/usr/local/sietium/mcu_upgrade.sh

# make sure every change write back to disk
sync
