/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef __DPTX_DRIVER_H__
#define __DPTX_DRIVER_H__

#include <linux/irq.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
#include <drm/drm_dp_helper.h>
#include <drm/drm_dp_mst_helper.h>
#else
#include <drm/display/drm_dp.h>
#include <drm/display/drm_dp_mst_helper.h>
#endif
#include <drm/drm_fixed.h>
//#include "drm_dp_helper_additions.h"

//#define GB02_PAL_DP
#define GB02MAC1318		0x100
#define GB02MAC1319		0x10
#define GB02MAC1320	0x9
#define GB02MAC1321	(9 * 4)
//#define DPTX_TYPE_C
#define GB02MAC1322

//DSC definitions coming from RC files of DSC C model
#define GB02MAC1323			8192
#define GB02MAC1324			6144
#define GB02MAC1325			512
#define GB02MAC1326		3
#define GB02MAC1327		5
#define GB02MAC1328		3
#define GB02MAC1329	9
#define GB02MAC1330		24
#define GB02MAC1331		GB02MAC1330
#define GB02MAC1332			3
#define	GB02MAC1333			12
#define GB02MAC1334			6
#define GB02MAC1335		11
#define GB02MAC1336		11
#define GB02MAC1337		3
#define GB02MAC1338		3
#define GB02MAC1339	3
#define GB02MAC1340		11

struct GB02STR140 {
	int minQP	:5;
	int maxQP	:5;
	int offset	:6;
};


// Rounding up to the nearest multiple of a number
#define ROUND_UP_TO_NEAREST(numToRound, mult) \
	((((numToRound+(mult)-1) / (mult)) * (mult)))


#include "gb_dp_avgen.h"
#include "gb_dp_reg.h"
#include "gb_dp_dbg.h"

struct dptx;

/* The max rate and lanes supported by the core */
#define GB02MAC1341 GB02MAC2181
#define GB02MAC1342 4

/* The default rate and lanes to use for link training */
#define GB02MAC1343 GB02MAC1341
#define GB02MAC1344 GB02MAC1342

/**
 * struct GB02STR141 - The link state.
 * @status: Holds the sink status register values.
 * @trained: True if the link is successfully trained.
 * @rate: The current rate that the link is trained at.
 * @lanes: The current number of lanes that the link is trained at.
 * @preemp_level: The pre-emphasis level used for each lane.
 * @vswing_level: The vswing level used for each lane.
 */
struct GB02STR141 {
	u8 status[DP_LINK_STATUS_SIZE];
	bool trained;
	u8 rate;
	u8 lanes;
	u8 preemp_level[4];
	u8 vswing_level[4];
};

enum established_timings {
	DMT_640x480_60hz,
	DMT_800x600_60hz,
	DMT_1024x768_60hz,
	NONE
};

/**
 * struct GB02STR142 - The aux state
 * @sts: The AUXSTS register contents.
 * @data: The AUX data register contents.
 * @event: Indicates an AUX event ocurred.
 * @abort: Indicates that the AUX transfer should be aborted.
 */
struct GB02STR142 {
	u32 sts;
	u32 data[4];
	atomic_t abort;
};

struct GB02STR143 {
	u8 HB0;
	u8 HB1;
	u8 HB2;
	u8 HB3;
} __packed;

struct GB02STR144 {
	u8 en;
	u32 payload[9];
	u8 blanking;
	u8 cont;
} __packed;

#define GB02MAC1351	0x331c1169
#define GB02MAC1352	10

struct GB02STR146 {
	u32 lsb;
	u32 msb;
};

struct GB02STR147 {
	u32 lsb;
	u32 msb;
};

struct GB02STR148 {
	struct GB02STR146 aksv;
	struct GB02STR147 dpk[40];
	u32 enc_key;
	u32 crc32;
	u8 auth_fail_count;
	int hdcp13_is_en;
	int hdcp22_is_en;
};

/**
 * struct dptx - The representation of the DP TX core
 * @mutex: dptx mutex
 * @base: Base address of the registers
 * @irq: IRQ number
 * @version: Contents of the IP_VERSION register
 * @max_rate: The maximum rate that the controller supports
 * @max_lanes: The maximum lane count that the controller supports
 * @dev: The struct device
 * @root: The debugfs root
 * @regset: The debugfs regset
 * @vparams: The video params to use
 * @aparams: The audio params to use
 * @hparams: The HDCP params to use
 * @waitq: The waitq
 * @shutdown: Signals that the driver is shutting down and that all
 *            operations should be aborted.
 * @c_connect: Signals that a HOT_PLUG or HOT_UNPLUG has occurred.
 * @sink_request: Signals the a HPD_IRQ has occurred.
 * @rx_caps: The sink's receiver capabilities.
 * @edid: The sink's EDID.
 * @sdp_list: The array of SDP elements
 * @aux: AUX channel state for performing an AUX transfer.
 * @link: The current link state.
 * @multipixel: Controls multipixel configuration. 0-Single, 1-Dual, 2-Quad.
 */
typedef struct dptx {
	struct mutex mutex; /* generic mutex for dptx */

	struct {
		u8 multipixel;
		u8 streams;
		bool gen2phy;
		bool dsc;
	} hwparams;

	void __iomem *base;
	int irq;

	u32 version;
	int index;
	u8 max_rate;
	u8 max_lanes;
	bool ycbcr420;
	u8 streams;

	bool mst;
	int active_mst_links;

	struct drm_dp_mst_topology_mgr mst_mgr;  //sahakyan

	bool cr_fail;// harutk need to remove

	u8 multipixel;
	u8 bstatus;

	bool ssc_en;
	bool fec;
	bool dsc;

	int vcp_id;
	u16 pbn;
	bool need_rad;
	bool logic_port;
	u8 rad_port;
	u8 port[2];

	bool dummy_dtds_present;
	enum established_timings selected_est_timing;

	struct device *dev;
	struct dentry *root;
	struct debugfs_regset32 *regset;

	struct GB02STR124 vparams;
	struct GB02STR119 aparams;
	struct GB02STR120 audio_desc;
	struct GB02STR148 hparams;

	wait_queue_head_t waitq;

	atomic_t shutdown;
	atomic_t c_connect;
	atomic_t sink_request;

	u8 rx_caps[GB02MAC1318];

	u8 *edid;
	u8 *edid_second;
#define GB02MAC1364 128

	struct GB02STR144 sdp_list[GB02MAC1319];
	struct GB02STR142 aux;
	struct GB02STR141 link;
	struct drm_device *drm_dev;
	struct delayed_work hotplug_work;
	bool is_edp;
	bool edid_getted;
}dptx_s;

/*
 * Core interface functions
 */
int GB02FUNC1065(struct dptx *dptx, bool is_edp);
void GB02FUNC1062(struct dptx *dptx);
bool GB02FUNC1058(struct dptx *dptx);
void GB02FUNC1050(struct dptx *dptx);
int GB02FUNC1054(struct dptx *dptx, bool sink_ssc);
bool GB02FUNC1052(struct dptx *dptx);
void GB02FUNC1060(struct dptx *dptx);

int GB02FUNC1074(struct dptx *dptx);
void GB02FUNC1042(struct dptx *dptx, u32 bits);
void GB02FUNC1047(struct dptx *dptx);
void GB02FUNC1048(struct dptx *dptx);

irqreturn_t dptx_irq(int irq, void *dev);
irqreturn_t dptx_threaded_irq(int irq, void *dev);

void GB02FUNC1038(struct dptx *dp);
void GB02FUNC1040(struct dptx *dp);
void GB02FUNC649(struct dptx *dptx);

/*
 * PHY IF Control
 */
void GB02FUNC1077(struct dptx *dptx, unsigned int num);
unsigned int GB02FUNC1075(struct dptx *dptx);
void GB02FUNC1082(struct dptx *dptx, unsigned int rate);
unsigned int GB02FUNC1087(struct dptx *dptx);
int GB02FUNC1090(struct dptx *dptx, unsigned int lanes);
void GB02FUNC1094(struct dptx *dptx,
			       unsigned int lane,
			       unsigned int pre,
			       unsigned int vsw);
void GB02FUNC1092(struct dptx *dptx,
	unsigned int lane, unsigned int level);
void GB02FUNC1103(struct dptx *dptx,
	unsigned int lane, unsigned int level);
void GB02FUNC1108(struct dptx *dptx,
			  unsigned int pattern);
void GB02FUNC1110(struct dptx *dptx, unsigned int lane, bool enable);

int GB02FUNC1112(unsigned int rate);
u8 GB02FUNC1114(unsigned int bw);

int GB02FUNC1549(struct dptx *dptx);
int GB02FUNC1538(struct dptx *dptx);
int GB02FUNC1552(struct dptx *dptx);
int GB02FUNC1141(struct dptx *dptx,
				   u8 rate,
				   u8 lanes);
/*
 * AUX Channel
 */

#define GB02MAC1389 2000

int GB02FUNC360(struct dptx *dptx,
	unsigned int device_addr, u8 *bytes, u32 len);

int GB02FUNC362(struct dptx *dptx,
	unsigned int device_addr);

int GB02FUNC366(struct dptx *dptx,
	unsigned int device_addr, u8 *bytes, u32 len);

int GB02FUNC378(struct dptx *dptx, u32 addr, u8 *byte);
int GB02FUNC380(struct dptx *dptx, u32 addr, u8 byte);

int GB02FUNC371(struct dptx *dptx,
				unsigned int reg_addr,
				u8 *bytes, u32 len);

int GB02FUNC376(struct dptx *dptx,
			       unsigned int reg_addr,
			       u8 *bytes, u32 len);

#ifndef GB02MAC1317
static inline int GB02FUNC1143(struct dptx *dptx, u32 addr, u8 *byte)
{
	return GB02FUNC378(dptx, addr, byte);
}

static inline int GB02FUNC1145(struct dptx *dptx, u32 addr, u8 byte)
{
	return GB02FUNC380(dptx, addr, byte);
}

static inline int GB02FUNC1146(struct dptx *dptx,
					    unsigned int reg_addr,
					    u8 *bytes, u32 len)
{
	return GB02FUNC371(dptx, reg_addr, bytes, len);
}

static inline int GB02FUNC1147(struct dptx *dptx,
					   unsigned int reg_addr,
					   u8 *bytes, u32 len)
{
	return GB02FUNC376(dptx, reg_addr, bytes, len);
}

#else
#define GB02FUNC1143(_dptx, _addr, _byteptr) ({		\
	int _ret = GB02FUNC378(_dptx, _addr, _byteptr);	\
	dptx_dbg(dptx, "%s: DPCD Read %s(0x%03x) = 0x%02x\n",	\
		 __func__, #_addr, _addr, *(_byteptr));		\
	_ret;							\
})

#define GB02FUNC1145(_dptx, _addr, _byte) ({			\
	int _ret;						\
	dptx_dbg(dptx, "%s: DPCD Write %s(0x%03x) = 0x%02x\n",	\
		 __func__, #_addr, _addr, _byte);		\
	_ret = GB02FUNC380(_dptx, _addr, _byte);		\
	_ret;							\
})

char *GB02FUNC1116(u8 *bytes, unsigned int n);

#define GB02FUNC1146(_dptx, _addr, _b, _len) ({	\
	int _ret;						\
	char *_str;						\
	_ret = GB02FUNC371(_dptx,		\
					   _addr, _b, _len);	\
	_str = GB02FUNC1116(_b, _len);				\
	dptx_dbg(dptx, "%s: Read %llu bytes from %s(0x%02x) = [ %s ]\n", \
		 __func__, (u64)_len, #_addr, _addr, _str);		\
	_ret;							\
})

#define GB02FUNC1147(_dptx, _addr, _b, _len) ({	\
	int _ret;						\
	char *_str = GB02FUNC1116(_b, _len);			\
	dptx_dbg(dptx, "%s: Writing %llu bytes to %s(0x%02x) = [ %s ]\n", \
		 __func__, (u64)_len, #_addr, _addr, _str);		\
	_ret = GB02FUNC376(_dptx, _addr, _b, _len); \
	_ret;							\
})

#endif

/*
 * Link training
 */
int GB02FUNC1554(struct dptx *dptx, u8 rate, u8 lanes);
int GB02FUNC1569(struct dptx *dptx);

/*
 * Register read and write functions
 */
static inline u32 GB02FUNC1149(struct dptx *dp, char const *func,
	int line, u32 offset)
{
	u32 data = readl(dp->base + offset);
#ifdef DPTX_DEBUG_REG
	dptx_dbg(dp, "%s:%d: READ: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base, offset, data);
#endif

	return data;
}
static inline u32 GB02FUNC1151(struct dptx *dp, char const *func,
	int line, u32 offset)
{
	u32 data = readl(dp->base + 0x20000 + offset);
#ifdef DPTX_DEBUG_REG
	dptx_dbg(dp, "%s:%d: dpphyREAD: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base, offset, data);
#endif

	return data;
}


static inline void GB02FUNC1153(struct dptx *dp, char const *func, int line,
	u32 offset, u32 data)
{
#ifdef DPTX_DEBUG_REG
	u32 rdata = 0;
	dptx_dbg(dp, "%s:%d: WRITE: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base, offset, data);
#endif
	writel(data, dp->base + offset);
#ifdef DPTX_DEBUG_REG
	rdata = GB02FUNC1149(dp, func, line, offset);
	if (rdata != data)
		gb_printf(KERN_ERR, "%s-%d:ERR,base:0x%p,addr:0x%x,data:0x%x,rdata:0x%x\n",
		 func, line, dp->base, offset, data, rdata);
#endif
}
static inline void GB02FUNC1155(struct dptx *dp, char const *func, int line,
	u32 offset, u32 data)
{
#ifdef DPTX_DEBUG_REG
	dptx_dbg(dp, "%s:%d: ~~~~~~~dpphy WRITE: base:0x%pK, addr=0x%05x data=0x%08x\n",
			func, line, dp->base + 0x20000, offset, data);
#endif
	writel(data, dp->base + 0x20000 + offset);
}

#define dptx_readl(_dptx, _offset) ({ \
	GB02FUNC1149(_dptx, __func__, __LINE__, _offset); \
})
#define dpphy_readl(_dptx, _offset) ({ \
	GB02FUNC1151(_dptx, __func__, __LINE__, _offset); \
})

#define dptx_writel(_dptx, _offset, _data) do { \
	GB02FUNC1153(_dptx, __func__, __LINE__, _offset, _data); \
} while (0)
#define dpphy_writel(_dptx, _offset, _data) do { \
	GB02FUNC1155(_dptx, __func__, __LINE__, _offset, _data); \
} while (0)

/*
 * Wait functions
 */
#define dptx_wait(_dptx, _cond, _timeout)				\
	({								\
		int __retval;						\
		__retval = wait_event_interruptible_timeout(		\
			_dptx->waitq,					\
			((_cond) || (atomic_read(&_dptx->shutdown))),	\
			msecs_to_jiffies(_timeout));			\
		if (atomic_read(&_dptx->shutdown)) {			\
			__retval = -ESHUTDOWN;				\
		}							\
		else if (!__retval) {					\
			__retval = -ETIMEDOUT;				\
		}							\
		__retval;						\
	})
#define dptx_link_wait(_dptx, _cond, _timeout)				\
	({								\
		int __retval;						\
		__retval = wait_event_interruptible_timeout(		\
			_dptx->waitq,					\
			(_cond),	\
			msecs_to_jiffies(_timeout));			\
		__retval;						\
	})

void dptx_notify(struct dptx *dptx);
void dptx_notify_shutdown(struct dptx *dptx);

/* Type C */
void GB02FUNC1399(struct dptx *dptx);

/*
 * Debugfs
 */
void dptx_debugfs_init(struct dptx *dptx);
void dptx_debugfs_exit(struct dptx *dptx);

void GB02FUNC421(struct dptx *dptx, struct GB02STR144 *data);
int dptx_configure_hdcp13(struct dptx *dptx);
void dptx_en_dis_hdcp13(struct dptx *dptx, u8 enable);
void dptx_init_hdcp_keys(struct dptx *dptx);

int dptx_en_hdcp22(struct dptx *dptx);
void dptx_en_dis_hdcp22(struct dptx *dptx, u8 enable);

int GB02FUNC1481(struct dptx *dptx, u8 enable);
int GB02FUNC1475(struct dptx *dptx, u8 enable);
int GB02FUNC1478(struct dptx *dptx, u8 enable);

struct dptx *dptx_get_handle(void);

int GB02FUNC351(struct dptx *dptx,
		      bool rw,
		      bool i2c,
		      u32 addr,
		      u8 *bytes,
		      unsigned int len);
int GB02FUNC360(struct dptx *dptx,
			     u32 device_addr,
			     u8 *bytes,
			     u32 len);
int GB02FUNC366(struct dptx *dptx,
			    u32 device_addr,
			    u8 *bytes,
			    u32 len);
int GB02FUNC331(struct dptx *dptx,
			bool rw,
			bool i2c,
			bool mot,
			bool addr_only,
			u32 addr,
			u8 *bytes,
			unsigned int len);
static inline int GB02FUNC1160(struct dptx *dptx, u32 frequency, int stream)
{
	return 0;
};
int GB02FUNC1161(struct dptx *dptx, u32 audio_clock_freq,
	u16 oversample_factor);
/*
 * HDCP2.2
 */
void dptx_write_hdcp22_test_keys(struct dptx *dptx);

//#define DPTX_COMBO_PHY
// To enable audio during hotplug
void GB02FUNC394(struct dptx *dptx);

void GB02FUNC396(struct dptx *dptx);
void GB02FUNC1521(struct dptx *dptx);
int GB02FUNC1546(struct dptx *dptx);
int GB02FUNC1547(struct dptx *dptx);
/* EDID Audio Data Block */
#define GB02MAC1461		1
#define GB02MAC1463		2
#define EDID_TAG_MASK		GENMASK(7, 5)
#define GB02MAC1465		5
#define EDID_SIZE_MASK		GENMASK(4, 0)
#define GB02MAC1468		0

/* Established timing blocks */
#define GB02MAC1469	BIT(0)
#define GB02MAC1472	BIT(1)
#define GB02MAC1474	BIT(2)
#define GB02MAC1476	BIT(3)
#define GB02MAC1478	BIT(4)
#define GB02MAC1480	BIT(5)
#define GB02MAC1482	BIT(6)
#define GB02MAC1484	BIT(7)

#define GB02MAC1486	BIT(0)
#define GB02MAC1488	BIT(1)
#define GB02MAC1490	BIT(2)
#define GB02MAC1492	BIT(3)
#define GB02MAC1494	BIT(4)
#define GB02MAC1496	BIT(5)
#define GB02MAC1498	BIT(6)
#define GB02MAC1500	BIT(7)

#define GB02MAC1502	BIT(7)

#endif
