/*
 * Copyright (C) Xi'an Sietium Electronics Co.Ltd
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include "common/gb_common.h"
#ifndef __GB_HDMAC_REGMAP_H__
#define __GB_HDMAC_REGMAP_H__

#define GB02MAC2713 GB02MAC503
#define GB02MAC2714 (10*MB)
#define GB02MAC2715 (GB02MAC503 + (GB02MAC491 - GB02MAC2714))

#define GB02MAC2716 0x210000000
#define GB02MAC2717 0x110000000

#define GB02MAC2718 48

#if GB02MAC2718 > 39
#define GB02MAC1194    0
#else
#define GB02MAC1194    1
#endif

#define GB02MAC1195 3


#define GB02MAC1193 512

#define GB02MAC310 (12)
#define GB02MAC309 (1 << GB02MAC310)
#define GB02MAC2719 (63)
#define GB02MAC2720 (11)

#define GB02MAC2721 0
#define GB02MAC2722 0

#if !GB02MAC2721
#define GB_DEBUG_PRINT_TRACE \
		"In file: " __FILE__
#if !GB02MAC2722
#define GB02MAC2723 __func__
#else
#define GB02MAC2723 ""
#endif
#else
#define GB_DEBUG_PRINT_TRACE ""
#endif

#define GB_DEBUG_ASSERT_OUT(trace, function, ...)\
		do { \
			gb_printf(KERN_ERR, "Botson<ASSERT>: %s function:%s ",\
			 trace, function);\
			gb_printf(KERN_ERR, __VA_ARGS__);\
			gb_printf(KERN_ERR, "\n");\
		} while (false)

#define GB_DEBUG_ASSERT(expr) \
	GB_DEBUG_ASSERT_MSG(expr, #expr)

#define GB_DEBUG_ASSERT_MSG(expr, ...) \
		do { \
			if (!(expr)) { \
				GB_DEBUG_ASSERT_OUT(GB_DEBUG_PRINT_TRACE,\
				 GB02MAC2723, __VA_ARGS__);\
				gb_printf(KERN_ERR, "Assert Error: inside %s %d\n",\
				 __func__, __LINE__); \
			} \
		} while (false)

enum gb_hdmac_mmu_fault_type {
	GB_MMU_FAULT_TYPE_UNKNOWN = 0,
	GB_MMU_FAULT_TYPE_PAGE,
	GB_MMU_FAULT_TYPE_BUS,
	GB_MMU_FAULT_TYPE_PAGE_UNEXPECTED,
	GB_MMU_FAULT_TYPE_BUS_UNEXPECTED
};

enum gb_hdmac_share_attr_bits {
	/* (1ULL << 8) bit is reserved */
	SHARE_BOTH_BITS = (2ULL << 8), /* inner and out shareable coherency */
	SHARE_INNER_BITS = (3ULL << 8) /* inner shareable coherency */
};

struct GB02STR211 {
	u64	transtab;
	u64	memattr;
	u64	transcfg;
};

struct GB02STR212 {
	int number;
	enum gb_hdmac_mmu_fault_type fault_type;
	bool protected_mode;
	u32 fault_status;
	u64 fault_addr;
	u64 fault_extra_addr;
	struct GB02STR211 current_setup;
};

struct GB02STR213 {
	pid_t tgid;
	pid_t pid;
	int as_nr;
	phys_addr_t pgd;
	struct GB02STR215 *mmu_mode;
	u64 fault_cpu_addr;
	u64 fault_gpu_addr;
};

struct GB02STR215 {
	void (*update)(struct GB02STR213 *gbctx);
	void (*get_as_setup)(struct GB02STR213 *gbctx,
			struct GB02STR211 * const setup);
	void (*disable_as)(struct GB02STR213 *gbctx);
	phys_addr_t (*GB02FUNC999)(u64 entry);
	int (*GB02FUNC1001)(u64 ate);
	int (*GB02FUNC1003)(u64 pte);
	void (*GB02FUNC1012)(u64 *entry, phys_addr_t phy, unsigned long flags);
	void (*GB02FUNC1014)(u64 *entry, phys_addr_t phy);
	void (*GB02FUNC1015)(u64 *entry);
};

/*
 * Maximum number of loops polling the GPU for a cache flush before
 * we assume it must have completed
 */
#define GB02MAC2727     100000
/*
 * Maximum number of loops polling the GPU for an AS command to complete
 * before we assume the GPU has hung
 */
#define GB02MAC2728     10


/* region flags */
#define GB02MAC1237              (1ul << 0)
/* CPU write access */
#define GB02MAC1239            (1ul << 1)
/* GPU write access */
#define GB02MAC1240            (1ul << 2)
/* No eXecute flag */
#define GB02MAC1242            (1ul << 3)
/* Is CPU cached? */
#define GB02MAC1244        (1ul << 4)
/* Is GPU cached? */
#define GB02MAC1245        (1ul << 5)

#define GB02MAC1248          (1ul << 6)
/* Can grow on pf? */
#define GB02MAC1249           (1ul << 7)

/* VA managed by us */
#define GB02MAC1251         (1ul << 8)

/* inner shareable coherency */
#define GB02MAC1253          (1ul << 9)
/* inner & outer shareable coherency */
#define GB02MAC1255        (1ul << 10)

/* Space for 4 different zones */
#define GB02MAC1257         (3ul << 11)
#define GB02MAC1258(x)           (((x) & 3) << 11)

/* GPU read access */
#define GB02MAC1259            (1ul<<13)
/* CPU read access */
#define GB02MAC1260            (1ul<<14)

/* Index of chosen MEMATTR for this region (0..7) */
#define GB02MAC1261      (7ul << 16)
#define GB02MAC1262(x)  (((x) & 7) << 16)
#define GB02MAC1263(x)  (((x) & GB02MAC1261) >> 16)

#define GB02MAC2730            (1ul << 19)

#define GB02MAC2731         (1ul << 20)

/* hdmA reg, just for read to debug */
#define GB02MAC2732		0x20
#define GB02MAC2733		0x24
#define GB02MAC2734		0x28
#define GB02MAC2735		0x2c
#define GB02MAC2736			0x1c
#define GB02MAC2737			0x00

#define GB02MAC2738		0x120
#define GB02MAC2739		0x124
#define GB02MAC2740		0x128
#define GB02MAC2741		0x12c
#define GB02MAC2742			0x11c
#define GB02MAC2743			0x100

#define GB02MAC2744(i)		(GB02MAC2732 + 0x200 * (i))
#define GB02MAC2745(i)		(GB02MAC2733 + 0x200 * (i))
#define GB02MAC2746(i)		(GB02MAC2734 + 0x200 * (i))
#define GB02MAC2747(i)		(GB02MAC2735 + 0x200 * (i))
#define GB02MAC2748(i)		(GB02MAC2736 + 0x200 * (i))
#define GB02MAC2749(i)		(GB02MAC2737 + 0x200 * (i))

#define GB02MAC2750(i)		(GB02MAC2738 + 0x200 * (i))
#define GB02MAC2751(i)		(GB02MAC2739 + 0x200 * (i))
#define GB02MAC2752(i)		(GB02MAC2740 + 0x200 * (i))
#define GB02MAC2753(i)		(GB02MAC2741 + 0x200 * (i))
#define GB02MAC2754(i)		(GB02MAC2742 + 0x200 * (i))
#define GB02MAC2755(i)		(GB02MAC2743 + 0x200 * (i))

#if 1
#define GB02MAC2756 0x84
#define GB02MAC2757 0x184
#define GB02MAC2758(i) (GB02MAC2756 + 0x200 * (i))
#define GB02MAC2759(i) (GB02MAC2757 + 0x200 * (i))
#else
#define GB02MAC2756 0x84
#define GB02MAC2757 0x284
#define GB02MAC2758(i) (GB02MAC2756 + 0x400 * (i))
#define GB02MAC2759(i) (GB02MAC2757 + 0x400 * (i))
#endif

/* hdmA reg end */

#define GB02MAC2762		4
#define GB02MAC2764	0xfffffffe
#define GB02MAC2766		0xfffffffd

/* hdmaC reg map start */
#define GB02MAC2769				0x2000
#define GB02MAC2771					(GB02MAC2769 + 0x20)
/* bit0:set 0 to softreset mmu; bit1:set 0 to softreset hdmac */
#define GB02MAC2772				(GB02MAC2769 + 0x24)
/* 0~7, totally 8 AS */
#define GB02MAC2773				(GB02MAC2769 + 0x30)
//#define DMA_CHANNEL				(GB02MAC2769 + 0x5c)
#define GB02MAC2775			(GB02MAC2769 + 0x48)
#define GB02MAC2776			(GB02MAC2769 + 0x4c)
#define GB02MAC2777			(GB02MAC2769 + 0x50)
#define GB02MAC2779			(GB02MAC2769 + 0x54)
#define GB02MAC2781				(GB02MAC2769 + 0x58)
#define GB02MAC2782				(GB02MAC2769 + 0x44)
/*
 * must be set at last... bit0 = 1:start dma;
 * bit31 = 1:dma read; bit31 = 0:dma_write
 */
#define GB02MAC2783				(GB02MAC2769 + 0x40)

#define GB02MAC2784			(GB02MAC2769 + 0x60)
#define GB02MAC2785		(GB02MAC2769 + 0x64)
#define GB02MAC2786		(GB02MAC2769 + 0x68)
#define GB02MAC2787			(GB02MAC2769 + 0x6c)
#define GB02MAC2789			(GB02MAC2769 + 0x70)

#define GB02MAC2790	(GB02MAC2769 + 0x80)
#define GB02MAC2791	(GB02MAC2769 + 0x84)
#define GB02MAC2792(r)    (GB02MAC2790 + (8 * (r)))
#define GB02MAC2794(r)    (GB02MAC2791 + (8 * (r)))

#define GB02MAC2796	(GB02MAC2769 + 0x100)
#define GB02MAC2798	(GB02MAC2769 + 0x104)
#define GB02MAC2800(r)    (GB02MAC2796 + (8 * (r)))
#define GB02MAC2802(r)    (GB02MAC2798 + (8 * (r)))

#define GB02MAC2803		(GB02MAC2769 + 0x180)
#define GB02MAC2804(r)     (GB02MAC2803 + (4 * (r)))

#define GB02MAC2805	(GB02MAC2769 + 0x1c0)

/* 0 dianping; 1maichong  */
#define GB02MAC2806		(GB02MAC2769 + 0x1c4)

#define GB02MAC2807		(GB02MAC2769 + 0x1c8)

/* enable hdmac to clear hdma irq */
#define GB02MAC2808	(GB02MAC2769 + 0x1cc)

/* hdmaC reg map end */

/* mmu reg map */
#define GB02MAC1569  0x0
#define GB02MAC1570(r)              (GB02MAC1569 + (r))

#define GB02MAC1575     0x0	/* Configuration registers for address space 0 */

#define MMU_AS_REG(n, r)        (GB02MAC1570(GB02MAC1575 + ((n) << 6)) + (r))

/* (RW) Translation Table Base Address for address space n, low word */
#define GB02MAC2809			0x00
/* (RW) Translation Table Base Address for address space n, high word */
#define GB02MAC2810			0x04
/* (RW) Memory attributes for address space n, low word. */
#define GB02MAC2811			0x08
/* (RW) Memory attributes for address space n, high word. */
#define GB02MAC2812			0x0C
/* (RW) Lock region address for address space n, low word */
#define GB02MAC2813			0x10
/* (RW) Lock region address for address space n, high word */
#define GB02MAC2814			0x14

#define GB02MAC1598	0x18 /* (WO) MMU command register for address space n */
/* (RO) MMU fault status register for address space n */
#define GB02MAC1600 0x1C
/* (RO) Fault Address for address space n, low word */
#define GB02MAC2815		0x20
/* (RO) Fault Address for address space n, high word */
#define GB02MAC2816		0x24
#define GB02MAC1602	0x28	/*(RO) Status flags for address space n*/


/* (RW) Translation table configuration for address space n, low word */
#define GB02MAC2817			0x30
/* (RW) Translation table configuration for address space n, high word */
#define GB02MAC2818			0x34
/* (RO) Secondary fault address for address space n, low word */
#define GB02MAC2819		0x38
/* (RO) Secondary fault address for address space n, high word */
#define GB02MAC2820		0x3C

#define GB02MAC1571		0x400	/* (RW) Raw interrupt status register */
#define GB02MAC1572		0x404	/* (WO) Interrupt clear register */
#define GB02MAC1573		0x408	/* (RW) Interrupt mask register */
#define GB02MAC1574		0x40C	/* (RO) Interrupt status register */

/* End Register Offsets */

/*
 * GB02MAC1571 register values. Values are valid also for
   GB02MAC1572, GB02MAC1573, GB02MAC1574 registers.
 */

#define GB02MAC1638   16

/* Macros returning a bitmask to retrieve page fault or bus error flags from
 * MMU registers
 */
#define GB02MAC1641(n)      (1UL << (n))
#define GB02MAC1643(n)       (1UL << ((n) + GB02MAC1638))

/*
 * Begin LPAE MMU TRANSTAB register values
 */
#define GB02MAC1646   0xfffff000
#define GB02MAC1648  (0u << 0)
#define GB02MAC1650  (1u << 1)
#define GB02MAC1653     (3u << 0)
#define GB02MAC1654        (1u << 2)
#define GB02MAC1655       (1u << 4)

#define GB02MAC1658      0x00000003

/*
 * Begin AARCH64 MMU TRANSTAB register values
 */
#define GB02MAC1218 40
#define GB02MAC1220 ((1ULL << GB02MAC1218) - (1ULL << 4))

/*
 * Begin MMU STATUS register values
 */
#define GB02MAC1666 0x01

#define GB02MAC1670                    (0x7<<3)
#define GB02MAC1674       (0x0<<3)
#define GB02MAC1676        (0x1<<3)
#define GB02MAC1679      (0x2<<3)
#define GB02MAC1681             (0x3<<3)

#define GB02MAC1684      (0x4<<3)
#define GB02MAC1687 (0x5<<3)

#define GB02MAC1704                  (0x3<<8)
#define GB02MAC1707                (0x0<<8)
#define GB02MAC1708                    (0x1<<8)
#define GB02MAC1711                  (0x2<<8)
#define GB02MAC1714                 (0x3<<8)

/*
 * Begin MMU TRANSCFG register values
 */

#define GB02MAC1209      0
#define GB02MAC1210    1
#define GB02MAC1212    2
#define GB02MAC1214  6
#define GB02MAC1216 8

#define GB02MAC1728        0xF


/*
 * Begin TRANSCFG register values
 */
#define GB02MAC1196 (3 << 24)
#define GB02MAC1197 (1 << 24)
#define GB02MAC1198 (2 << 24)

#define GB02MAC1735 ((3 << 28))
#define GB02MAC1737 (2 << 28)
#define GB02MAC1739 (3 << 28)

/* GB02MAC1598 register commands */
#define GB02MAC1760         0x00	/* NOP Operation */
/* Broadcasts the values in AS_TRANSTAB and ASn_MEMATTR to all MMUs */
#define GB02MAC1762      0x01
#define GB02MAC1764	0x02 /* Issue a lock region command to all MMUs */
#define GB02MAC1765 0x03 /* Issue a flush region command to all MMUs */
/*
 * Flush all L2 caches then issue a flush region command to all MMUs
 * (deprecated - only for use with T60x)
 */
#define GB02MAC1766 0x04
/* Flush all L2 caches then issue a flush region command to all MMUs */
#define GB02MAC1767    0x04
/*
 * Wait for memory accesses to complete, flush all the L1s cache then
 * flush all L2 caches then issue a flush region command to all MMUs
 */
#define GB02MAC2821   0x05

/* AS<n>_MEMATTR values: */
/* Use GPU implementation-defined caching policy. */
#define GB02MAC1199 0x88ull
/* The attribute set to force all resources to be cached. */
#define GB02MAC1200    0x8Full
/* Inner write-alloc cache setup, no outer caching */
#define GB02MAC1201           0x8Dull

/* Set to implementation defined, outer caching */
#define GB02MAC1202 0x88ull
/* Set to write back memory, outer caching */
#define GB02MAC1203       0x8Dull

/* HW implementation defined caching */
#define GB02MAC1204 0
/* Force cache on */
#define GB02MAC1205    1
/* Write-alloc */
#define GB02MAC1206           2
/* Outer coherent, inner implementation defined policy */
#define GB02MAC1207        3
/* Outer coherent, write alloc inner */
#define GB02MAC1208              4
struct GB02STR215 *GB02FUNC1811(void);
phys_addr_t GB02FUNC1770(struct GB02STR213 *gbctx);
void GB02FUNC1772(void);
phys_addr_t GB02FUNC1767(void);
void GB02FUNC1769(void);
void GB02FUNC1782(void);
int GB02FUNC1788(struct GB02STR213 *gbctx, size_t nr_pages,
	bool is_vram, phys_addr_t *phys, u64 *va);
int GB02FUNC1815(u64 size, phys_addr_t *src_phys, phys_addr_t *dst_phys,
	bool is_read, int channel_id, bool is_done,
	u64 abort_pg_num, u64 cpu_virt, u64 gpu_virt);
void GB02FUNC1813(void);
u32 GB02FUNC1753(u32 offset);
void GB02FUNC1751(u32 offset, u32 value);
void GB02FUNC1784(void);
void GB02FUNC1786(void);
void GB02FUNC1828(void);
u32 GB02FUNC1830(void);
#endif
